from pygryfsmart import GryfApi
from pygryfsmart.const import DriverFunctions, OutputActions

from .base import _GryfDevice

from typing import Any

class GryfThermostat(_GryfDevice):

    _t_state: float | None=None
    _o_state = False

    def __init__(
        self,
        name: str,
        id: int,
        pin: int,
        temperature_id: int,
        temperature_pin: int,
        differential: int,
        api: GryfApi,
    ):
        self._attributes = {
            "id out": id,
            "pin out": pin,
            "id temp": temperature_id,
            "pin temp": temperature_pin,
        }

        super().__init__(name,
                         id,
                         pin,
                         api)
        self._t_id = temperature_id
        self._t_pin = temperature_pin
        self._update_fun_ptr = None | Any
        self._target_temperature = 0
        self._enable = False
        self._differential = differential

    def subscribe(self , update_fun_ptr):
        self._api.subscribe(self._id , self._pin, DriverFunctions.OUTPUTS , self.update_out)
        self._api.subscribe(self._t_id , self._t_pin, DriverFunctions.TEMP , self.update_temperature)
        self._update_fun_ptr = update_fun_ptr

    async def update_temperature(self , state):
        self._t_state = state

        data = {
            DriverFunctions.TEMP: self._t_state,
            DriverFunctions.OUTPUTS: self._o_state
        }

        if self._enable and self._t_state:
            if self._t_state > self._target_temperature + self._differential:
                await self._api.set_out(self._id , self._pin , OutputActions.OFF)
            elif self._t_state < self._target_temperature - self._differential: 
                await self._api.set_out(self._id , self._pin , OutputActions.ON)

        await self._update_fun_ptr(data)

    def enable(self , enable):
        self._enable = enable

    async def set_target_temperature(self , temperature):
        self._target_temperature = temperature

        if self._enable:
            if self._t_state > self._target_temperature + self._differential:
                await self._api.set_out(self._id , self._pin , OutputActions.OFF)
            elif self._t_state < self._target_temperature - self._differential: 
                await self._api.set_out(self._id , self._pin , OutputActions.ON)

    async def update_out(self , state):
        self._o_state = state

        data = {
            DriverFunctions.TEMP: self._t_state,
            DriverFunctions.OUTPUTS: self._o_state
        }

        await self._update_fun_ptr(data)

    @property
    def name(self):
        return f"{self._name}"

    def change_differential(self, new_differential):
        self._differential = new_differential



from pygryfsmart import GryfApi
from pygryfsmart import GryfExpert

from .base import _GryfDevice

class _GryfExpert(_GryfDevice):

    _expert: GryfExpert

    def __init__(self , api: GryfApi) -> None:
        super().__init__("Gryf Expert" , 0 , 0 , api)

    async def start(self):

        self._expert = GryfExpert(self._api)
        await self._expert.start_server()

    async def stop(self):

        await self._expert.stop_server()
        self._attributes = {
        }



from pygryfsmart import GryfApi

from .base import _GryfDevice

class GryfReset(_GryfDevice):

    def __init__(
        self,
        api: GryfApi,
    ) -> None:
        self._attributes = {}

        super().__init__("Gryf RST",
                         0,
                         0,
                         api)

    @property
    def name(self):
        return "Gryf RST"

    async def reset_all(self):
        await self._api.reset(0 , True)

    async def reset_single_module(self , module):
        await self._api.reset(module , True)



from pygryfsmart import GryfApi
from pygryfsmart.const import ShutterStates
from pygryfsmart.device.classic_cover import GryfCover
from pygryfsmart.const import DriverActions, DriverFunctions

import asyncio
import logging

from .base import _GryfDevice

_LOGGER = logging.getLogger(__name__)

class GryfPCover(GryfCover):
    def __init__(self,
                 name: str,
                 id: int,
                 pin: int,
                 time: int,
                 api: GryfApi,
                 fun_ptr=None
                 ) -> None:

        super().__init__(
            name,
            id,
            pin,
            time,
            api,
            fun_ptr,
        )



"""File with GryfExpert server class."""

import asyncio
import logging

_LOGGER = logging.getLogger(__name__)

class GryfExpert:
    """GryfExpert class."""

    def __init__(
            self,
            api,
            host='127.0.0.1', 
            port=4210
        ) -> None:
        """Initialise GryfExpert class."""

        self.host = host
        self.port = port

        self.writer = None
        self.server = None
        self._enable = False
        self._api = api
        self.active_clients = set()
        self.server_task = None

    async def handle_client(
            self,
            reader, 
            writer
        ) -> None:
        """Handle new server client."""

        addr = writer.get_extra_info('peername')
        _LOGGER.debug(f"Connected with: {addr}")
        self.active_clients.add(writer)

        try:
            message = "\r\nConnected to RTI\r"
            writer.write(message.encode())
            await writer.drain()
            while True:
                data = await reader.read(1024)
                if not data:
                    _LOGGER.error(f"Client: {addr} stop connection")
                    break

                message = data.decode().strip() + "\n\r"
                _LOGGER.debug(f"message from {addr}: {message}")

                await self._api.send_data(f"{message}\n\r")
        except asyncio.CancelledError:
            _LOGGER.error(f"Client: {addr} stop connection")
        except Exception as e:
            _LOGGER.error(f"ocurred error from {addr}: {e}")
        finally:
            _LOGGER.debug(f"Closing connection with {addr}")
            writer.close()
            self.writer = None
            self.active_clients.remove(writer)
            await writer.wait_closed()
            return

    async def send_data(
            self,
            message
        ) -> None:
        """Sending data to GryfExpert."""

        if self.active_clients:
            try:
                message += "\n"
                for client in self.active_clients:
                    client.write(message.encode())
                    await client.drain()
            except Exception as e:
                _LOGGER.error(f"Unable to send message: {message}, error: {e}")
            finally:
                return

    async def stop_server(self) -> None:
        """Stop GryfExpert server."""
        if self.server:
            _LOGGER.debug("Stopping the server")
            self.server.close()
            await self.server.wait_closed()
            _LOGGER.debug("Server stopped")
        else:
            _LOGGER.error("Server is not running")

        self._enable = False

    async def start_server(self) -> None:
        """Start GryfExpert server."""
        self._api.subscribe_input_message(self.send_data)
        self.server = await asyncio.start_server(
            self.handle_client, self.host, self.port
        )

        addr = self.server.sockets[0].getsockname()
        _LOGGER.info(f"Server started on {addr}")

        self._enable = True
        self.server_task = asyncio.create_task(self.server.serve_forever())

    @property
    def enable(self):
        """Return is server enable."""
        return self._enable




from .output import GryfOutput
from .input import GryfInput
from .temperature import GryfTemperature
from .pwm import GryfPwm
from .input_line import GryfInputLine
from .output_line import GryfOutputLine
from .thermostat import GryfThermostat
from .classic_cover import GryfCover, GryfPercentCover
from .pcover import GryfPCover
from .reset import GryfReset
from .base import _GryfDevice

__all__ = ["GryfOutput",
           "GryfInput",
           "GryfTemperature",
           "GryfPwm",
           "GryfInputLine",
           "GryfOutputLine",
           "GryfThermostat",
           "GryfCover",
           "GryfPCover",
           "GryfReset",
           "GryfPercentCover"
           "_GryfDevice"
           ]



from enum import IntEnum

BAUDRATE = 115200
PORT = 4510

class OutputActions(IntEnum):
    """Enum with Output states."""

    ON = 1
    OFF = 2
    TOGGLE = 3

class ShutterStates(IntEnum):
    """Enum with Shutter states."""

    CLOSE = 1
    OPEN = 2
    STOP = 3
    STEP_MODE = 4

class KeyModes(IntEnum):
    "Enum with key modes."

    NO = 0
    NC = 1

class DriverFunctions():
    """Enum with driver functions."""

    INPUTS = "I"
    OUTPUTS = "O"
    PWM = "LED"
    COVER = "R"
    TEMP = "T"
    FIND = "AT+FIND"
    PONG = "PONG"
    PRESS_LONG = "PL"
    PRESS_SHORT = "PS"

class DriverActions():
    SET_OUT = "AT+SetOut"
    SET_COVER = "AT+SetRol"
    SET_PWM = "SetLED"
    PING = "PING"
    RESET = "AT+RST"
    SEARCH = "AT+Search"
    
    GET_IN_STATE = "AT+StanIN"
    GET_OUT_STATE = "AT+StanOUT"
    GET_SHUTTER_STATE = "AT+StanROL"

class ConfigurationFunctions():
    SET_PRESS_TIME = "AT+Key"

subscriptable_function = [
    DriverFunctions.INPUTS,
    DriverFunctions.OUTPUTS,
    DriverFunctions.TEMP,
    DriverFunctions.PWM,
    DriverFunctions.COVER,
]
#
# COMMAND_FUNCTION_IN = "I"
# COMMAND_FUNCTION_OUT = "O"
# COMMAND_FUNCTION_PWM = "LED"
# COMMAND_FUNCTION_COVER = "R"
# COMMAND_FUNCTION_FIND = "AT+FIND"
# COMMAND_FUNCTION_PONG = "PONG"
# COMMAND_FUNCTION_PRESS_SHORT = "PS"
# COMMAND_FUNCTION_PRESS_LONG = "PL"
# COMMAND_FUNCTION_TEMP = "T"
#
# COMMAND_FUNCTION_GET_IN_STATE = "AT+StanIN"
# COMMAND_FUNCTION_GET_OUT_STATE = "AT+StanOUT"
# COMMAND_FUNCTION_SET_OUT = "AT+SetOut"
# COMMAND_FUNCTION_SET_COVER = "AT+SetRol"
# COMMAND_FUNCTION_SET_PWM = "SetLED"
# COMMAND_FUNCTION_PING = "PING"
# COMMAND_FUNCTION_SET_PRESS_TIME = "AT+Key"
# COMMADN_FUNCTION_SEARCH_MODULE = "AT+Search"
# COMMAND_FUNCTION_RESET = "AT+RST"
#
CONF_ID = "id"
CONF_PIN = "pin"
CONF_PTR = "ptr"
CONF_FUNCTION = "fun"
CONF_TEMPERATURE = "temp"
CONF_OUT = "out"



import logging
import asyncio

_LOGGER = logging.getLogger(__name__)

class TCPClientHandler:
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._reader = None
        self._writer = None
        self._reconnect_interval = 5

    @property
    def port(self):
        return self._port


    async def open_connection(self):
        while True:
            try:
                self._reader, self._writer = await asyncio.open_connection(self._host, self._port)
                _LOGGER.info(f"Connection opened: {self._host}:{self._port}")
                return
            except Exception as e:
                _LOGGER.error(f"Error opening connection: {e}")
                await asyncio.sleep(self._reconnect_interval)

    async def close_connection(self):
        if self._writer:
            try:
                self._writer.close()
                await self._writer.wait_closed()
                _LOGGER.info("Connection closed.")
            except Exception as e:
                _LOGGER.warning(f"Error while closing connection: {e}")
            finally:
                self._reader = None
                self._writer = None
        else:
            _LOGGER.warning("No active connection to close.")


    async def send_data(self, data):
        if self._writer:
            try:
                self._writer.write(data.encode())
                await self._writer.drain()  
                _LOGGER.debug(f"Sent data: {data}")
            except Exception as e:
                _LOGGER.error(f"Error sending data: {e}")
                await self.reconnect()
        else:
            _LOGGER.warning("Cannot send data: No active connection.")

    async def read_data(self):
        if self._reader:
            try:
                data = await self._reader.read(1024)  
                if data:
                    decoded_data = data.decode().strip()
                    _LOGGER.debug(f"Received data: {decoded_data}")
                    return decoded_data
                else:
                    _LOGGER.warning("Connection closed by server.")
                    await self.reconnect()
                    return ""
            except Exception as e:
                _LOGGER.error(f"Error reading data: {e}")
                await self.reconnect()
                return ""
        else:
            _LOGGER.warning("Cannot read data: No active connection.")
            await self.reconnect()
            return ""

    async def reconnect(self):
        while True:
            try:
                await self.close_connection()
                await self.open_connection()
                return
            except Exception as e:
                _LOGGER.error(f"Reconnect failed: {e}")
                await asyncio.sleep(self._reconnect_interval)




import asyncio

from pygryfsmart import GryfApi
from pygryfsmart.const import ShutterStates, DriverFunctions

from .base import _GryfDevice

import logging

_LOGGER = logging.getLogger(__name__)

class Timer:
    _api: GryfApi
    _direction: int=0
    _time: int=0
    _defined: bool=False
    _shutter_position: float=0.0
    _percent_per_10_milisecounds: float
    _timer_on: bool=False
    _timer_callback_ptr = None
    _secound_timer_callback_ptr = None

    def __init__(self, api: GryfApi, time: int):
        self._api = api
        self._task = None

        self._percent_per_10_milisecounds = 100.0 / time

    def setup_timer(self, direction, time):
        self._direction = direction
        self._time = time

    def set_timer_callback(self, callback_ptr):
        self._timer_callback_ptr = callback_ptr

    def set_secound_timer_callback(self, callback_ptr):
        self._secound_timer_callback_ptr = callback_ptr

    async def __timer_task(self):
        time = self._time
        self._time = 0
        self._timer_on = True

        for i in range(time):
            if self._direction:
                self._shutter_position += self._percent_per_10_milisecounds
            else:
                self._shutter_position -= self._percent_per_10_milisecounds

            if self._timer_callback_ptr is not None:
                await self._timer_callback_ptr(self._shutter_position)

            if self._secound_timer_callback_ptr is not None:

                if self._shutter_position >= 0 and self._shutter_position <= 100:
                    await self._secound_timer_callback_ptr(self._shutter_position)
                elif self._shutter_position >= 0: 
                    await self._secound_timer_callback_ptr(0)
                else:
                    await self._secound_timer_callback_ptr(100)

            await asyncio.sleep(0.1)

        self._timer_on = False

    def start_timer(self):
        if self._time != 0:
            if self._task and not self._task.done():
                self._task.cancel()
            else:
                self._defined = True

            if self._direction:
                self._shutter_position += 7
            else:
                self._shutter_position -= 7

            self._task = asyncio.create_task(self.__timer_task())

    def stop_timer(self):
        if self._task and not self._task.done():
            self._task.cancel()

    @property
    def timer_enable(self):
        return self._timer_on

    @property
    def shutter_position(self):
        return self._shutter_position
    

class GryfCover(_GryfDevice):
    _is_opening: bool=False
    _is_closing: bool=False

    def __init__(
        self,
        name: str,
        id: int,
        pin: int,
        time: int,
        api: GryfApi,
        fun_ptr=None,
    ):
        super().__init__(name,
                         id,
                         pin,
                         api)

        self._time = time

        self._attributes = {
            "id": id,
            "pin": pin,
            "time": time
        }

        self._fun_ptr = fun_ptr
        self._api.subscribe(self._id , self._pin, DriverFunctions.COVER , self.__async_update)
        self._shutter_state = 0
        self._shutter_position = 0
        self._feedback_update = 0

        self._time_per_percent: float=time / 100.0
        self._timer = Timer(api, time)

    async def __async_update(self, state):
        self._feedback_update = 1
        self._timer.start_timer()

        if state == 0:
            self._timer.stop_timer()

        if state != 2 and state != 1:
            if self._shutter_state == 1:
                self._shutter_position = 1
            elif self._shutter_state == 2:
                self._shutter_position = 0

        self._shutter_state = state

        if self._fun_ptr:
            await self._fun_ptr(state)

    async def open_button_call(self):
        pass

    async def close_button_call(self):
        pass

    async def turn_on(self, time=None):
        if time is None:
            time = self._time

        self._timer.setup_timer(1, time)
        self._feedback_update = 0

        for k in range(5):
            await self._api.set_cover(self._id , self._pin , time , ShutterStates.OPEN)

            for i in range(10):
                if self._feedback_update:
                    return

                await asyncio.sleep((k + 1) * 0.03)

    async def turn_off(self, time=None):
        if time is None:
            time = self._time

        self._timer.setup_timer(0, time)
        self._feedback_update = 0

        for k in range(5):
            await self._api.set_cover(self._id , self._pin , time , ShutterStates.CLOSE)

            for i in range(10):
                if self._feedback_update:
                    return

                await asyncio.sleep((k + 1) * 0.03)

    async def toggle(self):
        if self._api._cover_toggle_metod:
            await self._api.set_cover(self._id , self._pin , self._time , ShutterStates.STEP_MODE)
        else:
            if self._shutter_position == 1:
                await self.turn_off()
            elif self._shutter_position == 0:
                await self.turn_on()

    async def stop(self, time: int | None=None):
        self._feedback_update = 0
        self._is_opening = False
        self._is_closing = False

        if self._timer._shutter_position < 1 or self._timer._shutter_position > 99:
            for k in range(5):
                if time is not None:
                    await self._api.set_cover(self._id , self._pin , time , ShutterStates.STOP)
                else:
                    await self._api.set_cover(self._id , self._pin , self._time , ShutterStates.STOP)

                for i in range(10):
                    if self._feedback_update:
                        self._timer.stop_timer()
                        return

                    await asyncio.sleep(k + 1 * 0.05)

    def load_last_shutter_position(self, position):
        self._shutter_position = position

        if position:
            self._current_cover_position = 100.0
        else:
            self._current_cover_position = 0.0

    def subscribe(self , update_fun_ptr):
        self._fun_ptr = update_fun_ptr

    @property
    def name(self):
        return f"{self._name}"

class GryfPercentCover(GryfCover):
    _expected_position: int=0
    _current_position: float=0.0
    _tilt_opening_time: int=20
    _tilt_position_callback_ptr = None
    _tilt_position: float=0.0
    _tilt_waiting_to_set = False

    def __init__(
        self,
        name: str,
        id: int,
        pin: int,
        time: int,
        api: GryfApi,
        fun_ptr=None,
    ) -> None:
        super().__init__(
            name,
            id,
            pin,
            time,
            api,
            fun_ptr,
        )

        self._timer.set_timer_callback(self.callback_from_timer)

    def set_tilt_opening_time(self, time: int):
        self._tilt_opening_time = time

    def set_tilt_position_callback_ptr(self, ptr):
        self._tilt_position_callback_ptr = ptr

    async def set_cover_tilt_position(self, position: int):
        if not self._is_closing and not self._is_opening:
            if self._tilt_position > position:
                time = (self._tilt_position - position) * (self._tilt_opening_time / 100.0)
                # cover_potition = self._current_position - ((time / self._time) * 100.0)
                await self.turn_on(time=time)

                # await self.set_cover_position(cover_potition)

            elif self._tilt_position < position:
                time = (position - self._tilt_position) * (self._tilt_opening_time / 100.0)
                # cover_potition = self._current_position + ((time / self._time) * 100.0)
                await self.turn_off(time=time)

                # await self.set_cover_position(cover_potition)
        else:
            self._tilt_waiting_to_set = True

    async def set_cover_position(self, expected_position):

        if expected_position > self._current_position:
            if not self._is_opening:
                await self.turn_on()

            self._is_opening = True
            self._is_closing = False

        elif expected_position < self._current_position:
            if not self._is_closing:
                await self.turn_off()

            self._is_opening = False
            self._is_closing = True

        self._expected_position = expected_position

    async def callback_from_timer(self, position: float):

        self._current_position = position

        if (
            (self._is_opening and self._current_position >= self._expected_position)
            or
            (self._is_closing and self._current_position <= self._expected_position)
            or
            (self._current_position <= 0 or self._current_position > 99)
        ):
            await self.stop()

            self._is_opening = False
            self._is_closing = False

        if self._tilt_position < 100 and self._is_opening:
            self._tilt_position += (100.0 / self._tilt_opening_time)

            if self._tilt_position_callback_ptr is not None:
                await self._tilt_position_callback_ptr(self._tilt_position)
        elif self._tilt_position > 0 and self._is_closing:
            self._tilt_position -= (100.0 / self._tilt_opening_time)

            if self._tilt_position_callback_ptr is not None:
                await self._tilt_position_callback_ptr(self._tilt_position)

    @property
    def tilt_is_waiting_to_set(self):
        if self._tilt_waiting_to_set:
            self._tilt_waiting_to_set = False

            return True

        return False



from pygryfsmart.const import OutputActions, DriverFunctions
from pygryfsmart import GryfApi

from .base import _GryfDevice

import asyncio

class GryfOutput(_GryfDevice):
    _state = 0
    _feedback_update = 0

    _update_fun_ptr = None

    def __init__(
        self,
        name: str,
        id: int,
        pin: int,
        api: GryfApi,
        update_fun_ptr=None,
    ):
        self._attributes = {
            "id": id,
            "pin": pin,
        }

        super().__init__(name,
                         id,
                         pin,
                         api)

        self._api.subscribe(self._id , self._pin , DriverFunctions.OUTPUTS , self.__async_update)
        
        self._update_fun_ptr = update_fun_ptr

    def subscribe(self , update_fun_ptr):
        self._update_fun_ptr = update_fun_ptr

    async def __async_update(self, state):
        self._state = state
        self._feedback_update = 1

        if self._update_fun_ptr:
            await self._update_fun_ptr(state)

    @property
    def name(self):
        return f"{self._name}"

    @property
    def state(self):
        return self._state

    async def turn_on(self):
        for k in range(10):

            self._feedback_update = 0
            await self._api.set_out(self._id, self._pin, OutputActions.ON)

            for i in range(10):

                if self._feedback_update:
                    break
                
                await asyncio.sleep(k * 0.01)
            
            if self._state == 1:
                break

    async def turn_off(self):
        for k in range(10):

            self._feedback_update = 0
            await self._api.set_out(self._id, self._pin, OutputActions.OFF)

            for i in range(10):

                if self._feedback_update:
                    break
                
                await asyncio.sleep(k * 0.01)
            
            if self._state == 0:
                break

    async def toggle(self):
        await self._api.set_out(self._id, self._pin, OutputActions.TOGGLE)



from pygryfsmart import GryfApi
from pygryfsmart.const import DriverFunctions

from .base import _GryfDevice

class GryfInput(_GryfDevice):

    def __init__(self,
                 name: str,
                 id: int,
                 pin: int,
                 api: GryfApi,
                 callback=None,
                 ) -> None:
        super().__init__(name,
                         id,
                         pin,
                         api)
        if callback is not None:
            self._api.subscribe(self._id , self._pin , DriverFunctions.INPUTS , callback)

    def subscribe(self , update_fun_ptr):
        self._api.subscribe(self._id , self._pin, DriverFunctions.INPUTS , update_fun_ptr)

    @property
    def name(self):
        return f"{self._name}"

